@@ -72,7 +72,10 @@ final_reduce(Reduce, {KVs, Reductions}) ->
 fold_reduce(#btree{root=Root}=Bt, Fun, Acc, Options) ->
     Dir = couch_util:get_value(dir, Options, fwd),
     StartKey = couch_util:get_value(start_key, Options),
-    EndKey = couch_util:get_value(end_key, Options),
+    EndKey = case couch_util:get_value(end_key_gt, Options) of
+        undefined -> couch_util:get_value(end_key, Options);
+        LastKey -> LastKey
+    end,
     KeyGroupFun = couch_util:get_value(key_group_fun, Options, fun(_,_) -> true end),
     {StartKey2, EndKey2} =
     case Dir of
@@ -191,7 +194,7 @@ query_modify(Bt, LookupKeys, InsertValues, RemoveKeys) ->
     {ok, NewRoot, Bt3} = complete_root(Bt2, KeyPointers),
     {ok, QueryResults, Bt3#btree{root=NewRoot}}.
 
-% for ordering different operatations with the same key.
+% for ordering different operations with the same key.
 % fetch < remove < insert
 op_order(fetch) -> 1;
 op_order(remove) -> 2;
