@@ -76,6 +76,7 @@ check_dbname(#server{dbname_regexp=RegExp}, DbName) ->
     nomatch ->
         case DbName of
             "_users" -> ok;
+            "_replicator" -> ok;
             _Else ->
                 {error, illegal_database_name}
             end;
@@ -167,7 +168,7 @@ all_databases() ->
             end,
             [list_to_binary(filename:rootname(RelativeFilename, ".couch")) | AccIn]
         end, []),
-    {ok, Filenames}.
+    {ok, lists:usort(Filenames)}.
 
 
 maybe_close_lru_db(#server{dbs_open=NumOpen, max_dbs_open=MaxOpen}=Server)
@@ -192,12 +193,7 @@ try_close_lru(StartTime) ->
         [{_, {opened, MainPid, LruTime}}] = ets:lookup(couch_dbs_by_name, DbName),
         case couch_db:is_idle(MainPid) of
         true ->
-            couch_util:shutdown_sync(MainPid),
-            true = ets:delete(couch_dbs_by_lru, LruTime),
-            true = ets:delete(couch_dbs_by_name, DbName),
-            true = ets:delete(couch_dbs_by_pid, MainPid),
-            true = ets:delete(couch_sys_dbs, DbName),
-            ok;
+            ok = shutdown_idle_db(DbName, MainPid, LruTime);
         false ->
             % this still has referrers. Go ahead and give it a current lru time
             % and try the next one in the table.
@@ -222,12 +218,22 @@ get_lru(LruTime) ->
         [{_, {opened, MainPid, _}}] = ets:lookup(couch_dbs_by_name, DbName),
         case couch_db:is_idle(MainPid) of
         true ->
-            LruTime;
+            NextLru = ets:next(couch_dbs_by_lru, LruTime),
+            ok = shutdown_idle_db(DbName, MainPid, LruTime),
+            get_lru(NextLru);
         false ->
             get_lru(ets:next(couch_dbs_by_lru, LruTime))
         end
     end.
 
+shutdown_idle_db(DbName, MainPid, LruTime) ->
+    couch_util:shutdown_sync(MainPid),
+    true = ets:delete(couch_dbs_by_lru, LruTime),
+    true = ets:delete(couch_dbs_by_name, DbName),
+    true = ets:delete(couch_dbs_by_pid, MainPid),
+    true = ets:delete(couch_sys_dbs, DbName),
+    ok.
+
 open_async(Server, From, DbName, Filepath, Options) ->
     Parent = self(),
     Opener = spawn_link(fun() ->
@@ -326,7 +332,7 @@ handle_call({delete, DbName, _Options}, _From, Server) ->
             couch_util:shutdown_sync(Pid),
             true = ets:delete(couch_dbs_by_name, DbName),
             true = ets:delete(couch_dbs_by_pid, Pid),
-            [gen_server:send_result(F, not_found) || F <- Froms],
+            [gen_server:reply(F, not_found) || F <- Froms],
             true;
         [{_, {opened, Pid, LruTime}}] ->
             couch_util:shutdown_sync(Pid),
