@@ -32,12 +32,7 @@ start_link(IniFiles) ->
     end.
 
 restart_core_server() ->
-    supervisor:terminate_child(couch_primary_services, couch_server),
-    supervisor:terminate_child(couch_secondary_services, stats_aggregator),
-    supervisor:terminate_child(couch_secondary_services, stats_collector),
-    supervisor:restart_child(couch_primary_services, couch_server),
-    supervisor:restart_child(couch_secondary_services, stats_collector),
-    supervisor:restart_child(couch_secondary_services, stats_aggregator).
+    init:restart().
 
 couch_config_start_link_wrapper(IniFiles, FirstConfigPid) ->
     case is_process_alive(FirstConfigPid) of
@@ -124,9 +119,24 @@ start_server(IniFiles) ->
     unlink(ConfigPid),
 
     Ip = couch_config:get("httpd", "bind_address"),
-    Port = mochiweb_socket_server:get(couch_httpd, port),
     io:format("Apache CouchDB has started. Time to relax.~n"),
-    ?LOG_INFO("Apache CouchDB has started on http://~s:~w/", [Ip, Port]),
+    Uris = [get_uri(Name, Ip) || Name <- [couch_httpd, https]],
+    [begin
+        case Uri of
+            undefined -> ok;
+            Uri -> ?LOG_INFO("Apache CouchDB has started on ~s", [Uri])
+        end
+    end
+    || Uri <- Uris],
+    case couch_config:get("couchdb", "uri_file", null) of 
+    null -> ok;
+    UriFile ->
+        Lines = [begin case Uri of
+            undefined -> [];
+            Uri -> io_lib:format("~s~n", [Uri])
+            end end || Uri <- Uris],
+        file:write_file(UriFile, Lines)
+    end,
 
     {ok, Pid}.
 
@@ -174,7 +184,7 @@ start_secondary_services() ->
             {list_to_atom(Name),
                 {Module, Fun, Args},
                 permanent,
-                brutal_kill,
+                1000,
                 worker,
                 [Module]}
         end
@@ -189,3 +199,22 @@ stop() ->
 
 init(ChildSpecs) ->
     {ok, ChildSpecs}.
+
+get_uri(Name, Ip) ->
+    case get_port(Name) of
+        undefined ->
+            undefined;
+        Port ->
+            io_lib:format("~s://~s:~w/", [get_scheme(Name), Ip, Port])
+    end.
+
+get_scheme(couch_httpd) -> "http";
+get_scheme(https) -> "https".
+
+get_port(Name) ->
+    try
+        mochiweb_socket_server:get(Name, port)
+    catch
+        exit:{noproc, _}->
+            undefined
+    end.
